==> tty-client-L0.07.tar.gz: tty-client-L0.07/read-line.c <==
/*
 *  client/read-line.c
 *		Routines to read one line from file.
 *		Automatically (re)allocates enough space for
 *		line.  These routines are system dependant, and
 *		will probably need to be fixed when porting.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */



#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <s-string.h>
#include <zmalloc.h>

#include "error.h"

#include "read-line.h"



#define Export


/**************************************
*     Local function declarations     *
**************************************/


static void
erase_backward_char (FILE   * terminal);

static void
move_cursor_backwards (int	no_of_chars,
		       FILE   * terminal   );

static void
erase_to_end_of_line (FILE   * terminal);

static void
alert_user (FILE   * terminal);




/*  ==================================================================  */
/*			    Exported functions				*/


/*
 *  Read a string from a terminal into a String with some fancy
 *  options and some editing, including delete character, delete
 *  word and kill line.  Automatically allocates all necessary
 *  storage.
 *  Parameters:
 *  -  INFILE:	the file to read the string from.
 *  -  ECHOFILE: the file to which echoing is done.
 *  -  LINE:	pointer to the String where the result should be placed.
 *		This String must be initialized to something useful,
 *		i e EMPTY_STRING or a zmalloc():ed area.
 *  -  MAXLEN: maximum number of characters to allow in input.
 *		Excess characters are ignored.  -1 means no limit.
 *  -  QUIT_CHARS: a NUL-terminated string of characters that can
 *		be used to end the input.  Typically contains the
 *		newline character ('\n').  The character used to
 *		exit, will be part of the resulting string.
 *  -  INSERT_CHARS: a NUL-terminated string of characters that
 *		can build up the resulting String.  NULL means
 *		normal printable characters including tab.  ""
 *		means all characters without special meaning.
 *  -  ECHO_CHAR: the caracter that is to be echoed. NUL means no
 *		echoing is done, and EOF means echo the read
 *		characters.
 *  Returns values:
 *	    0	Something went wrong - probably out of memory
 *	  EOF	End of file reached (includes Ctrl-D)
 *	Other	The character that caused the exit
 *
 *  Even if input_line() returns FAILURE, the value of *LINE is
 *  valid, and can be zfree():ed or zrealloc():ed, though the contents
 *  is probably just gibberish.
 */
Export  int
input_line (FILE		* infile,
	    FILE		* echofile,
	    String		* line,
	    String_size		  maxlen,
	    char		* quit_chars,
	    char		* insert_chars,
	    int			  echo_char)

{
    char		* work_string;
    char		* temp_string; 		/* Temporary storage	*/
    String_size		  charno;		/* Current character no */
    String_size		  alloced_size;		/* Allocated size	*/
    String_size		  i;			/* Loop counter		*/
    char		  c;			/* Read character	*/
    Bool		  end;			/* Time to exit?	*/
    int			  exit_value;		/* Value to return	*/
    struct  sgttyb	  tty_param;
    struct  tchars	  tty_chars;
    struct  ltchars	  more_chars;

#define FUNCTION "input_line()"

    #define		  chunk_size	32



    /* Fetch tty settings */
    if (ioctl (fileno(infile), TIOCGETP, (caddr_t) &tty_param) == -1)
    {
	/* This *shouldn't* fail, since init_terminal()
	 * didn't abort...				*/
    	perror("Oops!  You shouldn't see this.  Anyway, here's the error:\n"
	       "Fetching tty parameters: ");
	fatal (CLIENT_ERROR_TOO_boring, "Fetching tty parameters", errno);
    }

    /* Get special chars, specially EOF-char */
    if (ioctl (fileno(infile), TIOCGETC, (caddr_t) &tty_chars) == -1)
    {
	fatal (CLIENT_ERROR_TOO_boring,
			    "Fetching EOF character", errno);
    }

    /* Get more special chars, specially reprint, word erase and
     * literal next.						*/
    if (ioctl (fileno(infile), TIOCGLTC, (caddr_t) &more_chars) == -1)
    {
	fatal (CLIENT_ERROR_TOO_boring,
			    "Fetching more characters", errno);
    }
    
    /* Free the old line, and allocate new work-string */
    work_string = zrealloc (line->string, alloced_size = chunk_size);
    if (work_string == NULL)
    {
	/*  We seem to have a problem, Dave.  We need more core.  */
	return  0;
    }

    end = FALSE;
    charno = 0;
    while (! end)
    {
	c = getc (infile);
	if (c == tty_param.sg_erase)		/* Delete one char */
	{
	    if (charno != 0)
	    {
		if (echo_char != '\000')
		    erase_backward_char (echofile);
		charno--;
	    }
	}
	else if	(c == tty_param.sg_kill)	/* Kill entire line */
	{
	    if (echo_char != '\000')
	    {
		move_cursor_backwards (charno, echofile);
		erase_to_end_of_line (echofile);
	    }
	    charno = 0;
	}
	else if (c == more_chars.t_werasc)	/* Kill word */
	{
	    if (charno != 0)
	    {
		while ((charno > 0) && (isspace (work_string[charno - 1])))
		{
		    if (echo_char != '\000')
			erase_backward_char (echofile);
		    charno--;
		}
		while ((charno > 0) && (! isspace (work_string[charno - 1])))
		{
		    if (echo_char != '\000')
			erase_backward_char (echofile);
		    charno--;
		}
	    }
	}
	else if (c == more_chars.t_rprntc)	/* Reprint line */
	{
	    if (echo_char != '\000')
	    {
		putc('\n', echofile);
		for ( i = 0 ;  i < charno ;  i++ )
		    putc(work_string[i], echofile);
	    }
	}
	else if (c == tty_chars.t_eofc || c == EOF) /* End of file */
	{
	    exit_value = EOF;
	    end = TRUE;
	}
	/* 			* Character to be inserted in string */
	else if (  (strchr (quit_chars, c) != NULL)
		 || (  (insert_chars == NULL
			? (! iscntrl(c) || c == '\t')
			: (  *insert_chars == '\000'
			   || strchr (insert_chars, c) != NULL))
		     && ((maxlen == -1) || (charno < maxlen-1))))
	{
	    /* Do we need a larger work-string? */
	    if (charno >= alloced_size)
	    {
		temp_string = zrealloc (work_string,
					alloced_size += chunk_size);
		if (temp_string == NULL)
		{
		    /* Memory is tired - it is exhausted */
		    end = TRUE;
		    exit_value = 0;
		}
		work_string = temp_string;
	    }

	    work_string [charno++] = c;

	    if	(strchr (quit_chars, c) != NULL) /* Return or equivalent */
	    {
		exit_value = c;
		end = TRUE;
	    }
	    else
	    {
		if (echo_char == -1)
		    putc (c, echofile);
		else if (echo_char != '\000')
		    putc (echo_char, echofile);
	    }
	}
	else					/* Bad input */
	{
	    alert_user (echofile);
	}

    }	/**  while (! end)  **/

    line->string = work_string;
    line->len = charno;

    return  exit_value;
#undef FUNCTION
}




Export  int
read_line (String	* line)
{
    int		  retval;

    retval = input_line (stdin, stdout, line, -1, "\n", NULL, EOF);
    if (retval == '\n')
	line->len--;
    return  retval;
}



Export  int
read_password (String	* line)
{
    int		  retval;

    retval = input_line (stdin, stdout, line, -1, "\n", NULL, '*');
    if (retval == '\n')
	line->len--;
    return  retval;
}



/*************************************
*     Local function definitions     *
*************************************/


static  void
erase_backward_char (FILE   * terminal)
{
    fputs("\b \b", terminal);
}


static  void
move_cursor_backwards (int	  no_of_chars,
		       FILE	* terminal    )
{
    for ( ;  no_of_chars-- ; )
	putc('\b', terminal);
}


static  void
alert_user (FILE   * terminal)
{
    putc('\a', terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static void
erase_to_end_of_line (FILE   * terminal)
{
    fputs("\033[0K", terminal);
}



#if  0		/* Don't warn about unused functions */

static  void
cursor_at_start_of_line (FILE	* terminal)
{
    putc('\r', terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
erase_entire_line (FILE	  * terminal)
{
    fputs("\033[2K\015", terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
swedish_font (FILE   * terminal)
{
    fputs("\033<\033(B\033)1\033)H\016", terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
american_font (FILE   * terminal)
{
    fputs("\033<\033(B\033)1\033)G\017", terminal);
}

#endif	/* 0 */

==> tty-client-L0.08.tar.gz: tty-client-L0.08/read-line.c <==
/*
 *  client/read-line.c
 *		Routines to read one line from file.
 *		Automatically (re)allocates enough space for
 *		line.  These routines are system dependant, and
 *		will probably need to be fixed when porting.
 *
 *
 *  Copyright (C) 1990	Lysator Computer Club,
 *			Linkoping University,  Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author:	Thomas Bellman
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 *  email:	Bellman@Lysator.LiU.SE
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL opinions,
 *  and does NOT, repeat NOT, represent any official standpoint of Lysator,
 *  even if so stated.
 */



#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <s-string.h>
#include <zmalloc.h>

#include "error.h"

#include "read-line.h"



#define Export


/**************************************
*     Local function declarations     *
**************************************/


static void
erase_backward_char (FILE   * terminal);

static void
move_cursor_backwards (int	no_of_chars,
		       FILE   * terminal   );

static void
erase_to_end_of_line (FILE   * terminal);

static void
alert_user (FILE   * terminal);




/*  ==================================================================  */
/*			    Exported functions				*/


/*
 *  Read a string from a terminal into a String with some fancy
 *  options and some editing, including delete character, delete
 *  word and kill line.  Automatically allocates all necessary
 *  storage.
 *  Parameters:
 *  -  INFILE:	the file to read the string from.
 *  -  ECHOFILE: the file to which echoing is done.
 *  -  LINE:	pointer to the String where the result should be placed.
 *		This String must be initialized to something useful,
 *		i e EMPTY_STRING or a zmalloc():ed area.
 *  -  MAXLEN: maximum number of characters to allow in input.
 *		Excess characters are ignored.  -1 means no limit.
 *		-64 means that linebreak is done at next space after
 *		64 read characters.
 *  -  QUIT_CHARS: a NUL-terminated string of characters that can
 *		be used to end the input.  Typically contains the
 *		newline character ('\n').  The character used to
 *		exit, will be part of the resulting string.
 *  -  INSERT_CHARS: a NUL-terminated string of characters that
 *		can build up the resulting String.  NULL means
 *		normal printable characters including tab.  ""
 *		means all characters without special meaning.
 *  -  ECHO_CHAR: the caracter that is to be echoed. NUL means no
 *		echoing is done, and EOF means echo the read
 *		characters.
 *  Returns values:
 *	    0	Something went wrong - probably out of memory
 *	  EOF	End of file reached (includes Ctrl-D)
 *	Other	The character that caused the exit
 *
 *  Even if input_line() returns FAILURE, the value of *LINE is
 *  valid, and can be zfree():ed or zrealloc():ed, though the contents
 *  is probably just gibberish.
 */
Export  int
input_line (FILE		* infile,
	    FILE		* echofile,
	    String		* line,
	    String_size		  maxlen,
	    char		* quit_chars,
	    char		* insert_chars,
	    int			  echo_char)

{
    char		* work_string;
    char		* temp_string; 		/* Temporary storage	*/
    String_size		  charno;		/* Current character no */
    String_size		  alloced_size;		/* Allocated size	*/
    String_size		  i;			/* Loop counter		*/
    String_size		  linelen;              /* The length of the line */
    char		  c;			/* Read character	*/
    Bool		  end;			/* Time to exit?	*/
    int			  exit_value;		/* Value to return	*/
    struct  sgttyb	  tty_param;
    struct  tchars	  tty_chars;
    struct  ltchars	  more_chars;

#define FUNCTION "input_line()"

    #define		  chunk_size	32



    /* Fetch tty settings */
    if (ioctl (fileno(infile), TIOCGETP, (caddr_t) &tty_param) == -1)
    {
	/* This *shouldn't* fail, since init_terminal()
	 * didn't abort...				*/
    	perror("Oops!  You shouldn't see this.  Anyway, here's the error:\n"
	       "Fetching tty parameters: ");
	fatal (CLIENT_ERROR_TOO_boring, "Fetching tty parameters", errno);
    }

    /* Get special chars, specially EOF-char */
    if (ioctl (fileno(infile), TIOCGETC, (caddr_t) &tty_chars) == -1)
    {
	fatal (CLIENT_ERROR_TOO_boring,
			    "Fetching EOF character", errno);
    }

    /* Get more special chars, specially reprint, word erase and
     * literal next.						*/
    if (ioctl (fileno(infile), TIOCGLTC, (caddr_t) &more_chars) == -1)
    {
	fatal (CLIENT_ERROR_TOO_boring,
			    "Fetching more characters", errno);
    }
    
    /* Free the old line, and allocate new work-string */
    work_string = zrealloc (line->string, alloced_size = chunk_size);
    if (work_string == NULL)
    {
	/*  We seem to have a problem, Dave.  We need more core.  */
	return  0;
    }

    end = FALSE;
    charno = 0;
    linelen = 0;
    while (! end)
    {
	c = getc (infile);
	if (c == tty_param.sg_erase)		/* Delete one char */
	{
	    if (charno != 0)
	    {
		if (echo_char != '\000')
		    erase_backward_char (echofile);
		charno--;
		linelen--;
	    }
	}
	else if	(c == tty_param.sg_kill)	/* Kill entire line */
	{
	    if (echo_char != '\000')
	    {
		move_cursor_backwards (charno, echofile);
		erase_to_end_of_line (echofile);
		linelen = 0;
	    }
	    charno = 0;
	}
	else if (c == more_chars.t_werasc)	/* Kill word */
	{
	    if (charno != 0)
	    {
		while ((charno > 0) && (isspace (work_string[charno - 1])))
		{
		    if (echo_char != '\000')
			erase_backward_char (echofile);
		    charno--;
		    linelen--;
		}
		while ((charno > 0) && (! isspace (work_string[charno - 1])))
		{
		    if (echo_char != '\000')
			erase_backward_char (echofile);
		    charno--;
		    linelen--;
		}
	    }
	}
	else if (c == more_chars.t_rprntc)	/* Reprint line */
	{
	    if (echo_char != '\000')
	    {
		putc('\n', echofile);
		for ( i = 0 ;  i < charno ;  i++ )
		    putc(work_string[i], echofile);
	    }
	}
	else if (c == tty_chars.t_eofc || c == EOF) /* End of file */
	{
	    exit_value = EOF;
	    end = TRUE;
	}
	/* 			* Character to be inserted in string */
	else if (  (strchr (quit_chars, c) != NULL)
		 || (  (insert_chars == NULL
			? (! iscntrl(c) || c == '\t')
			: (  *insert_chars == '\000'
			   || strchr (insert_chars, c) != NULL))
		     && ((maxlen < 0) || (charno < maxlen-1))))
	{
	    /* Do we need a larger work-string? */
	    if (charno >= alloced_size)
	    {
		temp_string = zrealloc (work_string,
					alloced_size += chunk_size);
		if (temp_string == NULL)
		{
		    /* Memory is tired - it is exhausted */
		    end = TRUE;
		    exit_value = 0;
		}
		work_string = temp_string;
	    }

	    linelen++;		/* BUG should handle tabs. */

	    if (maxlen < -1)	/* Break at whitespace enabled. */
	    {
		if (linelen  > -maxlen
		    && strchr(" \t", c)) /* Whitespace. */ {
		    c = '\n';
		    linelen = 0;
		}
	    }

	    work_string [charno++] = c;

	    if	(strchr (quit_chars, c) != NULL) /* Return or equivalent */
	    {
		exit_value = c;
		end = TRUE;
	    }
	    else
	    {
		if (echo_char == -1)
		    putc (c, echofile);
		else if (echo_char != '\000')
		    putc (echo_char, echofile);
	    }
	}
	else					/* Bad input */
	{
	    alert_user (echofile);
	}

    }	/**  while (! end)  **/

    line->string = work_string;
    line->len = charno;

    return  exit_value;
#undef FUNCTION
}




Export  int
read_line (String	* line)
{
    int		  retval;

    retval = input_line (stdin, stdout, line, -1, "\n", NULL, EOF);
    if (retval == '\n')
	line->len--;
    return  retval;
}



Export  int
read_password (String	* line)
{
    int		  retval;

    retval = input_line (stdin, stdout, line, -1, "\n", NULL, '*');
    if (retval == '\n')
	line->len--;
    return  retval;
}



/*************************************
*     Local function definitions     *
*************************************/


static  void
erase_backward_char (FILE   * terminal)
{
    fputs("\b \b", terminal);
}


static  void
move_cursor_backwards (int	  no_of_chars,
		       FILE	* terminal    )
{
    for ( ;  no_of_chars-- ; )
	putc('\b', terminal);
}


static  void
alert_user (FILE   * terminal)
{
    putc('\a', terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static void
erase_to_end_of_line (FILE   * terminal)
{
    fputs("\033[0K", terminal);
}



#if  0		/* Don't warn about unused functions */

static  void
cursor_at_start_of_line (FILE	* terminal)
{
    putc('\r', terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
erase_entire_line (FILE	  * terminal)
{
    fputs("\033[2K\015", terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
swedish_font (FILE   * terminal)
{
    fputs("\033<\033(B\033)1\033)H\016", terminal);
}


/*  Hardcoded for vt100 compatible terminals... */
static  void
american_font (FILE   * terminal)
{
    fputs("\033<\033(B\033)1\033)G\017", terminal);
}

#endif	/* 0 */
